/* Smile
 * 10-06-2006
 * www.assert.i8.com
 * mail-to: db05@bk.ru
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define SOURCE  0
#define TARGET  1
#define COMMAND 2
#define MINARGC 2
#define MAXARGC 3


static FILE* f_source = NULL;
static FILE* f_target = NULL;
static long total = 0;
static char tmpbuf[1000];
static char cmdbuf[sizeof(tmpbuf)];
static int exitcode = EXIT_SUCCESS;

void
Usage()
{
	puts("PatchIt version 0.2 ("__DATE__" "__TIME__")\n");
	puts(" usage: patchit source_file target_file [command]");
	exit(EXIT_FAILURE);
}


void
Cleanup()
{
	if(f_source != NULL)
		fclose(f_source);

	if(f_target != NULL)
		fclose(f_target);
}

void
Error(const char* msg)
{
	perror(msg);
	Cleanup();
	exit(EXIT_FAILURE);
}


int
ParseError(const char* msg)
{
	fprintf(stderr,"parse error: %s\n -> %s\n",msg,cmdbuf);
	Cleanup();
	exit(EXIT_FAILURE);
	//return((exitcode = EXIT_FAILURE));
}


int
HexStr2Long(const char* p, long* value)
{
	char * errchar;
	*value = strtoul(p,&errchar,16);
	return(*errchar == '\0');
}


long
fsize(FILE* f)
{
	long size = -1;
	fpos_t pos;

	if(fgetpos(f,&pos) != 0)
		return(-1);

	if(fseek(f,0,SEEK_END) != 0)
		return(-1);

	if((size = ftell(f)) == -1)
		return(-1);

	if(fsetpos(f,&pos) != 0)
		return(-1);

	return(size);
}


void
LogChanges(long s_pos, long t_pos, long size)
{
	total += size;
	printf(" %08X %08X %07X\n",s_pos,t_pos,size);
}

void
LogResult()
{
	printf(" %8C %8C[%07X]\n",' ',' ',total);
}

int
MakeChanges(long s_pos, long t_pos, long size, void* buffer)
{
	long s_size, t_size;

	s_size = fsize(f_source);
	t_size = fsize(f_target);

	if(s_size == -1 || t_size == -1)
		Error("fsize");

	if((s_pos+size > s_size) || fseek(f_source,s_pos,SEEK_SET) != 0)
		return(ParseError("out of source file"));

	if((t_pos+size > t_size) || fseek(f_target,t_pos,SEEK_SET) != 0)
		return(ParseError("out of target file"));

	fread(buffer,size,1,f_source);
	if(ferror(f_source))
		Error("fread");

	fwrite(buffer,size,1,f_target);
	if(ferror(f_target))
		Error("fwrite");

	LogChanges(s_pos,t_pos,size);
	return(EXIT_SUCCESS);
}


int
ApplyCommand(int argc, char** argv)
{
	void* buffer;
	long s_pos=0, t_pos=0, size=0;

	if(!HexStr2Long(argv[0],&s_pos)
	|| !HexStr2Long(argv[1],&t_pos)) {
		return(ParseError("bad offset value"));
	}

	if(argc >= 3) {
		if(!HexStr2Long(argv[2],&size))
			return(ParseError("bad size value"));
	} else {
		size = fsize(f_source);
	}

	if((buffer = malloc(size)) == NULL)
		return(ParseError("no enougth memory or size too lage"));

	MakeChanges(s_pos,t_pos,size,buffer);

	free(buffer);
	return(exitcode);
}


void
Execute(char pass_char)
{
	int n;
	char *p, *p2;
	char* argv[MAXARGC] = {NULL,NULL,NULL};

	// check string
	if(pass_char == 0) {
		p = tmpbuf;
	} else {
		p = strchr(tmpbuf,pass_char);
		if(p == NULL)
			return;
		p++;
	}

	// split arguments
	for(n=0; n<MAXARGC; p++) {
		if(isxdigit(*p)) {
			for(p2=p; isxdigit(*p2); p2++);
			argv[n++] = p;
			p2[0] = '\0';
			p = p2;
			continue;
		}

		if(*p == '\n' || *p == '\r' || *p == '\0') {
			if(n == 0) return;
			break;
		}

		if(isspace(*p))
			continue;

		ParseError("unexpected characters");
		return;
	}

	if(n < MINARGC) {
		ParseError("bad command format");
		return;
	}
	
	ApplyCommand(n,argv);
}


int
main(int argc, char** argv)
{
	argc--;
	argv++;
	if(argc < COMMAND)
		Usage();

	if((f_source = fopen(argv[SOURCE],"rb")) == NULL)
		Error(argv[SOURCE]);

	if((f_target = fopen(argv[TARGET],"rb+")) == NULL)
		Error(argv[TARGET]);

	// Read commands
	if(argc <= COMMAND) {
		// from stdin
		while(fgets(tmpbuf,sizeof(tmpbuf),stdin) != NULL) {
			strncpy(cmdbuf,tmpbuf,sizeof(tmpbuf));
			Execute('#');
		}
	} else {
		// from command line
		for(int i=COMMAND,n=0; i<argc; i++)
			n += _snprintf(tmpbuf+n,sizeof(tmpbuf)-n,"%s ",argv[i]);

		strncpy(cmdbuf,tmpbuf,sizeof(tmpbuf));
		Execute(0);
	}

	LogResult();
	fclose(f_source);
	fclose(f_target);
	return(exitcode);
}
